# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# invenio-administration is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""
from collections import namedtuple

import pytest
from flask_security import login_user
from invenio_access.models import ActionRoles, Role
from invenio_access.permissions import superuser_access
from invenio_accounts.testutils import login_user_via_session
from invenio_admin.permissions import action_admin_access
from invenio_app.factory import create_app as _create_app
from invenio_records_resources.resources import RecordResource
from invenio_records_resources.services import RecordService
from mock_module.administration.mock import MockView
from mock_module.config import ServiceConfig
from mock_module.resource import MockResource


@pytest.fixture(scope="module")
def celery_config():
    """Override pytest-invenio fixture.

    TODO: Remove this fixture if you add Celery support.
    """
    return {}


@pytest.fixture(scope="module")
def extra_entry_points():
    """Register extra entry point."""
    return {
        "invenio_administration.views": [
            "mock_module = mock_module.administration.mock:MockView",
            "mock_module = mock_module.administration.mock:MockViewAlternate",
        ]
    }


@pytest.fixture
def admin_role_need(db):
    """Store 1 role with 'superuser-access' ActionNeed.

    WHY: This is needed because expansion of ActionNeed is
         done on the basis of a User/Role being associated with that Need.
         If no User/Role is associated with that Need (in the DB), the
         permission is expanded to an empty list.
    """
    role = Role(name="admin-access", id="admin-access")
    db.session.add(role)

    action_role = ActionRoles.create(action=action_admin_access, role=role)
    db.session.add(action_role)

    db.session.commit()

    return action_role.need


@pytest.fixture()
def admin(UserFixture, app, db, admin_role_need):
    """Admin user for requests."""
    u = UserFixture(
        email="admin@inveniosoftware.org",
        password="admin",
    )
    u.create(app, db)

    datastore = app.extensions["security"].datastore
    _, role = datastore._prepare_role_modify_args(u.user, "admin-access")

    datastore.add_role_to_user(u.user, role)
    db.session.commit()
    return u


@pytest.fixture
def superuser_identity(admin, superuser_role_need):
    """Superuser identity fixture."""
    identity = admin.identity
    identity.provides.add(superuser_role_need)
    return identity


@pytest.fixture
def superuser_role_need(db):
    """Store 1 role with 'superuser-access' ActionNeed.

    WHY: This is needed because expansion of ActionNeed is
         done on the basis of a User/Role being associated with that Need.
         If no User/Role is associated with that Need (in the DB), the
         permission is expanded to an empty list.
    """
    role = Role(name="superuser-access")
    db.session.add(role)

    action_role = ActionRoles.create(action=superuser_access, role=role)
    db.session.add(action_role)

    db.session.commit()

    return action_role.need


RunningApp = namedtuple(
    "RunningApp",
    [
        "app",
        "superuser_identity",
        "location",
        "cache",
    ],
)


@pytest.fixture(scope="module")
def create_app(instance_path, entry_points):
    """Application factory fixture."""
    return _create_app


@pytest.fixture
def running_app(
    app,
    superuser_identity,
    location,
    cache,
):
    """This fixture provides an app with the typically needed db data loaded.

    All of these fixtures are often needed together, so collecting them
    under a semantic umbrella makes sense.
    """
    return RunningApp(
        app,
        superuser_identity,
        location,
        cache,
    )


@pytest.fixture
def test_app(running_app):
    """Get current app."""
    return running_app.app


@pytest.fixture
def client_with_login(client, admin):
    """Log in a user to the client."""
    user = admin.user
    login_user(user)
    login_user_via_session(client, email=user.email)
    return client


@pytest.fixture
def current_admin_menu(test_app):
    """Current Admin flask menu instance."""
    return test_app.extensions["menu"]


@pytest.fixture
def current_admin_ext(test_app):
    """Current invenio-administration extension."""
    return test_app.extensions["invenio-administration"]


@pytest.fixture
def mock_service():
    """Creates a mock service."""
    service = RecordService(ServiceConfig)
    return service


@pytest.fixture
def mock_resource(mock_service):
    """Creates a mock resource."""
    return RecordResource(MockResource, mock_service)


@pytest.fixture(scope="function", autouse=True)
def register_mock_extension(test_app, mock_resource, mock_extension_name):
    """Creates a register a mock extension. Feature is used automatically all tests."""

    class MockExtension:
        def __init__(self, app=None):
            """Extension initialization."""
            if app:
                self.init_app(app)

        def init_app(self, app):
            app.extensions[mock_extension_name] = self
            #  Adds resource's config name as a class attribute.
            self.__dict__[MockView.resource_config] = mock_resource

    ext = MockExtension(test_app)
    return ext


@pytest.fixture
def mock_extension(test_app, mock_extension_name):
    """Creates a mock extension."""
    return test_app.extensions.get(mock_extension_name)


@pytest.fixture
def mock_extension_name():
    """Retrieves mock extension name."""
    return "mock-module"


@pytest.fixture
def current_admin_core(current_admin_ext):
    """Current Admin instance."""
    return current_admin_ext.administration


@pytest.fixture
def current_admin_bluepint(current_admin_core):
    """Retrieve admin blueprint."""
    return current_admin_core.blueprint



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Mock resource created for testing."""

from flask_resources import ResourceConfig


class MockResource(ResourceConfig):
    """Creates a mock resource for testing."""

    blueprint_name = "mocks"
    url_prefix = "/mocks"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022-2024 CERN.
# Copyright (C) 2023 KTH Royal Institute of Technology.
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio administration marshmallow utils module."""

from invenio_vocabularies.services.schema import (
    BaseVocabularySchema,
    ContribVocabularyRelationSchema,
    VocabularyRelationSchema,
)
from marshmallow import fields
from marshmallow_utils import fields as invenio_fields
from marshmallow_utils.fields import EDTFDateString, EDTFDateTimeString

vocabulary_schemas = [
    ContribVocabularyRelationSchema,
    BaseVocabularySchema,
    VocabularyRelationSchema,
]

custom_mapping = {
    # marshmallow
    fields.Str: "string",
    fields.Integer: "integer",
    fields.List: "array",
    fields.Dict: "object",
    fields.Url: "string",
    fields.String: "string",
    fields.DateTime: "datetime",
    fields.Float: "float",
    fields.Boolean: "bool",
    fields.Raw: "dict",
    fields.UUID: "uuid",
    fields.Time: "time",
    fields.Date: "date",
    fields.TimeDelta: "timedelta",
    fields.Decimal: "decimal",
    # TODO: This is needed for the is_current_user in the administration
    # of the user resources. It might be better to implement also a handler for this.
    # See https://github.com/inveniosoftware/invenio-administration/issues/174
    # For the moment it is skipped when we call `jsonify_schema`
    fields.Method: None,
    # invenio fields
    invenio_fields.SanitizedUnicode: "string",
    invenio_fields.links.Links: "array",
    invenio_fields.links.Link: "string",
    invenio_fields.tzdatetime.TZDateTime: "datetime",
    invenio_fields.sanitizedhtml.SanitizedHTML: "string",
    invenio_fields.isodate.ISODateString: "date",
    invenio_fields.url.URL: "string",
    EDTFDateString: "date",
    EDTFDateTimeString: "datetime",
}


def find_type_in_mapping(field_type, custom_mapping):
    """
    Find a field type by traversing the inheritance chain.

    Args:
        field_type (Type): The field type to be searched.
        custom_mapping (dict): The mapping of types.

    Returns:
        The mapped value for the found field type.

    Raises:
        KeyError: If the field type is not found in the mapping.
    """
    current_type = field_type
    while current_type:
        if current_type in custom_mapping:
            return custom_mapping[current_type]
        current_type = current_type.__base__

    raise KeyError(f"Unrecognized field type: {field_type}")


def jsonify_schema(schema):
    """Marshmallow schema to dict."""
    schema_dict = {}

    for field, field_type in schema.fields.items():
        is_links = isinstance(field_type, invenio_fields.links.Links)
        # skip `fields.Method`
        is_method = isinstance(field_type, fields.Method)

        if is_links or is_method:
            continue

        is_read_only = field_type.dump_only
        is_create_only = (
            field_type.metadata["create_only"]
            if "create_only" in field_type.metadata
            else False
        )

        field_type_name = field_type.__class__
        is_required = field_type.required

        nested_field = isinstance(field_type, fields.Nested)
        list_field = isinstance(field_type, fields.List)

        schema_dict[field] = {
            "required": is_required,
            "readOnly": is_read_only,
            "title": (
                field_type.metadata["title"] if "title" in field_type.metadata else None
            ),
            "createOnly": is_create_only,
            "metadata": field_type.metadata,
        }

        if nested_field:
            if any([isinstance(field_type.schema, x) for x in vocabulary_schemas]):
                schema_type = "vocabulary"
            else:
                schema_type = "object"

            schema_dict[field].update(
                {
                    "type": schema_type,
                    "properties": jsonify_schema(field_type.schema),
                }
            )
        elif list_field and isinstance(field_type.inner, fields.Nested):
            # list of objects (vocabularies or nested)
            schema_dict[field].update(
                {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": jsonify_schema(field_type.inner.schema),
                    },
                }
            )
        elif list_field and not isinstance(field_type.inner, fields.Nested):
            # list of plain types
            schema_dict[field].update(
                {
                    "type": "array",
                    "items": {
                        "type": find_type_in_mapping(
                            field_type.inner.__class__, custom_mapping
                        )
                    },
                }
            )
        else:
            try:
                field_type_mapping = find_type_in_mapping(
                    field_type_name, custom_mapping
                )
                schema_dict[field].update(
                    {
                        "type": field_type_mapping,
                    }
                )
            except KeyError:
                raise Exception(f"Unrecognised schema field {field}: {field_type_name}")
    return schema_dict



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Mock view for testing."""

from invenio_administration.views.base import (
    AdminResourceDetailView,
    AdminResourceListView,
)


class MockView(AdminResourceListView):
    """Creates a mock AdminView for testing.

    resource_config must match the mock resource's name
    """

    name = "mock"
    category = "Test category"
    url = "mocked_details_url"
    resource_config = "mocks"


class MockViewAlternate(AdminResourceListView):
    """Creates a mock AdminView for testing.

    resource_config must match the mock resource's name
    """

    name = "mock alternate"
    category = "Test category"
    resource_config = "mocks"
    # url is None to force the name to be used as url



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# InvenioAdministration is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio-Administration Permissions Generators."""

from invenio_records_permissions.generators import Generator

from invenio_administration.permissions import administration_access_action


class Administration(Generator):
    """Allows administration-access."""

    def __init__(self):
        """Constructor."""
        super(Administration, self).__init__()

    def needs(self, **kwargs):
        """Enabling Needs."""
        return [administration_access_action]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio admin extension."""

import importlib_metadata

from . import config
from .admin import Administration
from .views.base import AdminResourceBaseView, AdminView


class InvenioAdministration:
    """Invenio extension."""

    def __init__(self, app=None, entry_point_group="invenio_administration.views"):
        """Extension initialization."""
        self.entry_point_group = entry_point_group

        self.administration = None
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize application."""
        self.init_config(app)
        self.administration = Administration(
            app,
            name=app.config["ADMINISTRATION_APPNAME"],
            base_template=app.config["ADMINISTRATION_BASE_TEMPLATE"],
        )
        if self.entry_point_group:
            self.load_entry_point_group(app)
        app.extensions["invenio-administration"] = self

    def load_entry_point_group(self, app):
        """Load admin interface views from entry point group."""
        entrypoints = set(importlib_metadata.entry_points(group=self.entry_point_group))
        for ep in entrypoints:
            entry_point = self._load_entry_point(ep)
            entrypoint_path = ep.value

            extension_name_from_path = self._extract_extension_name(entrypoint_path)
            extension_name_from_view = entry_point.extension_name
            # fallback to extracted extension_name if property not set on view
            extension_name = extension_name_from_view or extension_name_from_path

            self.register_view(entry_point, extension_name, app)
        app.register_blueprint(self.administration.blueprint)

    def _load_entry_point(self, entry_point):
        """Loads one entry point. Validates whether its view is an AdminView."""
        ep = entry_point.load()
        if not issubclass(ep, AdminView):
            raise TypeError(f"View class must be of type {AdminView.__name__}")
        return ep

    def register_view(self, view_class, extension_name, app, *args, **kwargs):
        """Register an admin view on this admin instance.

        :param view_class: The view class name passed to the view factory.
        :param extension_name: The name of the extension associated with the view.
        :param args: Positional arguments for view class.
        :param kwargs: Keyword arguments to view class.
        """
        view_instance = view_class(
            extension_name=extension_name, admin=self.administration, *args, **kwargs
        )
        view = view_class.as_view(
            view_class.name,
            extension_name=extension_name,
            admin=self.administration,
            *args,
            **kwargs,
        )
        self.administration.add_view(view, view_instance, *args, **kwargs)
        if issubclass(view_class, AdminResourceBaseView):
            self.register_resource(app, view_class, extension_name)

    @staticmethod
    def register_resource(app, view_class, extension_name):
        """Set views schema."""

        @app.before_first_request
        def register_view_resource():
            if view_class.resource_config:
                view_class.set_resource(extension_name=extension_name)
            if view_class.schema:
                view_class.set_schema(extension_name=extension_name)

    @staticmethod
    def _extract_extension_name(entrypoint_path):
        name = entrypoint_path.split(".")[0]
        normalized_name = name.replace("_", "-")
        return normalized_name

    @staticmethod
    def init_config(app):
        """Initialize configuration.

        :param app: The Flask application.
        """
        # Set default configuration
        for k in dir(config):
            if k.startswith("ADMINISTRATION_"):
                app.config.setdefault(k, getattr(config, k))
